/*
 * Copyright (C) 2010 EPFL (Ecole Polytechnique fédérale de Lausanne)
 * Transport and Mobility Laboratory (Transp-OR)
 *
 * This file is part of COBI.
 *
 * COBI is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * COBI is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * COBI; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA.
 *
 * Author: Javier Cruz-Mota
 *         javier.cruz{AT}epfl.ch
 *         http://transp-or2.epfl.ch/pagesPerso/javier.php
 *
 */

#include "Cobi.h"

// Constructor
Cobi::Cobi(QWidget *parent)
    : QMainWindow(parent)
{
	ui.setupUi(this);

	videoLoaded        = false;
	playing            = false;
	paused             = false;
	lastOriginalFrame  = false;
	lastProcessedFrame = false;
	savingOutput       = false;
	timingIsOK	   = false;

	bufferSize    = ReadBufferWindow::DEFAULTBUFFERSIZE;
	outputQuality = OutputQualityWindow::DEFAULTQUALITY;

	loadedVideoData = new VideoData();
	refreshTimer    = new QTimer(this);

	originalImage.clear();
	processInputImage.clear();
	processedImage.clear();
	outputImage.clear();

	readingThread    = NULL;
	processingThread = NULL;
	writingThread    = NULL;

	ui.lineEdit->setDisabled(true);
}

// Destructor
Cobi::~Cobi()
{
	if(readingThread!=NULL)
	{
		emit stopReadingThread();
		readingThread->wait();
		delete readingThread;
	}

	if(processingThread!=NULL)
	{
		emit stopProcessingThread();
		processingThread->wait();
		delete processingThread;
	}

	delete loadedVideoData;

	if(refreshTimer->isActive())
	{
		refreshTimer->stop();
	}
	delete refreshTimer;

	if(savingOutput && writingThread!=NULL)
	{
		writingThread->wait();
		delete writingThread;
	}
	originalImage.clear();
	processInputImage.clear();
	processedImage.clear();
	outputImage.clear();

	ui.playSurface->makeCurrent();
		glDeleteTextures(1, &originalTextureName);
	ui.playSurface->doneCurrent();
	ui.playSurface->updateGL();

	ui.playSurface2->makeCurrent();
		glDeleteTextures(1, &processedTextureName);
	ui.playSurface2->doneCurrent();
	ui.playSurface2->updateGL();

	ui.verticalLayout->removeWidget(ui.playSurface);
	ui.horizontalLayout->removeWidget(ui.playButton);
	ui.horizontalLayout->removeItem(ui.horizontalSpacer);
	ui.horizontalLayout->removeWidget(ui.pauseButton);
	ui.horizontalLayout->removeItem(ui.horizontalSpacer_2);
	ui.horizontalLayout->removeWidget(ui.stopButton);
	ui.menubar->clear();

	delete ui.lineEdit;
	delete ui.statusbar;
	delete ui.menuSetup;
	delete ui.menubar;
	delete ui.label;
	delete ui.checkBox;
	delete ui.frame;
	delete ui.outputLabel;
	delete ui.inputLabel;
	delete ui.playSurface2;
	delete ui.stopButton;
	delete ui.horizontalSpacer_2;
	delete ui.pauseButton;
	delete ui.horizontalSpacer;
	delete ui.playButton;
	delete ui.horizontalLayout;
	delete ui.playSurface;
	delete ui.verticalLayout;
	delete ui.verticalLayoutWidget;
	delete ui.centralwidget;
	delete ui.actionOutputQuality;
	delete ui.actionReadBuffer;
	delete ui.actionLoadVideoFile;
	delete ui.actionQuit;
	delete ui.actionInfo;
}

// Returns true if a video has been loaded
bool Cobi::isVideoLoaded()
{
	return videoLoaded;
}

// "File->Info" action
void Cobi::on_actionInfo_activated()
{
	QString message("<CENTER><B>COBI v0.3 Copyright (C) 2010 EPFL</B></CENTER>"
					"<P><CENTER>Javier Cruz-Mota<BR>Transport and Mobility Laboratory<BR>EPFL</CENTER></P>"
					"<P>This program comes with ABSOLUTELY NO WARRANTY. "
					"This is free software, and you are welcome to redistribute it under certain conditions. "
					"For further details, see COPYING</P>");
	InfoWindow iwin(this, message);
    iwin.exec();
}

// "File->Quit" action
void Cobi::on_actionQuit_activated()
{
	this->close();
}

// "File->Load Video File" action
void Cobi::on_actionLoadVideoFile_activated()
{
    LoadWindow lwin(this);
    if (lwin.exec()) {
    	videoPath = lwin.getVideoPath();
    	QString videoFile = lwin.getVideoFile();
        if (!videoPath.isEmpty() && !videoFile.isEmpty()) {

        	QString statusMessage("Loading ");
        	statusMessage.append(videoFile);
        	ui.statusbar->showMessage(statusMessage,5000);

			videoLoaded = this->loadVideo();
        	if (videoLoaded) {
        		statusMessage.append(" DONE!");
        	}
        	else {
        		statusMessage.append(" ERROR!");
        	}
    		ui.statusbar->showMessage(statusMessage,5000);
        }
    }
}

// "Setup->Reading Buffer Size" action
void Cobi::on_actionReadBuffer_activated(void)
{
	ReadBufferWindow rbwin(this,bufferSize);
	if(rbwin.exec())
	{
		bufferSize = rbwin.getValue();
	}
}

// "Setup->Output Quality" action
void Cobi::on_actionOutputQuality_activated(void)
{
	OutputQualityWindow oqwin(this,outputQuality);
	if(oqwin.exec())
	{
		outputQuality = oqwin.getValue();
	}
}

// Called when there is a change in the state of the check box for saving
void Cobi::on_checkBox_stateChanged(void)
{
    QFont font;
    font.setPointSize(12);
    font.setStrikeOut(false);
    font.setKerning(false);

	if(ui.checkBox->isChecked())
	{
		font.setItalic(false);
		ui.lineEdit->setFont(font);
		ui.lineEdit->setEnabled(true);
		ui.lineEdit->setReadOnly(false);
		savingOutput = true;
	}
	else
	{
		font.setItalic(true);
		ui.lineEdit->setFont(font);
		ui.lineEdit->setDisabled(true);
		ui.lineEdit->setReadOnly(true);
		savingOutput = false;
	}
}

// Loads a video file
bool Cobi::loadVideo(void)
{
	if(loadedVideoData->LoadData(videoPath.toStdString().data()))
	{
		// If a video is being played, stop it and kill threads
		if(playing || paused)
		{
			playing = false;
			paused = false;

			emit stopReadingThread();
			readingThread->wait();
			delete readingThread;
			readingThread = NULL;

			emit stopProcessingThread();
			processingThread->wait();
			delete processingThread;
			processingThread = NULL;
			if(savingOutput)
			{
				writingThread->wait();
				delete writingThread;
				writingThread = NULL;
			}
		}

		unsigned char *tmpData;
		tmpData = (unsigned char *)new GLuint[((loadedVideoData->getWidth()*loadedVideoData->getHeight())* 4 * sizeof(unsigned char))];

		// prepare OpenGL surfaces
		ui.playSurface->makeCurrent();
			glEnable(GL_TEXTURE_2D);
			glGenTextures(1, &originalTextureName);
			glBindTexture(GL_TEXTURE_2D, originalTextureName);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glShadeModel(GL_SMOOTH);
			glDisable(GL_NORMALIZE);
			glDisable(GL_LIGHTING);
			glDisable(GL_BLEND);
			glDisable(GL_DEPTH_TEST);
			glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glViewport(0, 0, 480, 360);
			gluOrtho2D(0.0, 1.0, 0.0, 1.0);
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
						 loadedVideoData->getWidth(),
						 loadedVideoData->getHeight(),
						 0, GL_RGBA, GL_UNSIGNED_BYTE, tmpData);
			glFlush();
		ui.playSurface->doneCurrent();
		ui.playSurface->updateGL();

		ui.playSurface2->makeCurrent();
			glEnable(GL_TEXTURE_2D);
			glGenTextures(1, &processedTextureName);
			glBindTexture(GL_TEXTURE_2D, processedTextureName);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glPixelStorei(GL_PACK_ALIGNMENT, 1);
			glShadeModel(GL_SMOOTH);
			glDisable(GL_NORMALIZE);
			glDisable(GL_LIGHTING);
			glDisable(GL_BLEND);
			glDisable(GL_DEPTH_TEST);
			glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glViewport(0, 0, 480, 360);
			gluOrtho2D(0.0, 1.0, 0.0, 1.0);
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
						 loadedVideoData->getWidth(),
						 loadedVideoData->getHeight(),
						 0, GL_RGBA, GL_UNSIGNED_BYTE, tmpData);
			glFlush();
		ui.playSurface2->doneCurrent();
		ui.playSurface2->updateGL();

		delete [] tmpData;

		period = loadedVideoData->getPeriod();
		fps = 1000/period;

		originalImage.clear();
		processInputImage.clear();
		processedImage.clear();

		return true;
	}
	return false;
}

// Called when play button is clicked
void Cobi::on_playButton_clicked()
{
	if (videoLoaded && !playing && !paused) //Video is not being played
	{
		// Disable check box and line edit
		QFont font;
		font.setPointSize(12);
		font.setStrikeOut(false);
		font.setKerning(false);
		font.setItalic(true);
		ui.lineEdit->setFont(font);
		ui.lineEdit->setReadOnly(true);
		ui.lineEdit->setDisabled(true);
		ui.checkBox->setDisabled(true);


		// Launch reading thread and connect signals with slots
		playing = true;
		readingThread = new ReadVideoThread(this,loadedVideoData, &originalImage, &processInputImage, &bufferSize,
											&originalImageMutex, &processInputImageMutex);

		QObject::connect(readingThread, SIGNAL(newPushedImage(void)), this, SLOT(newReceivedImage(void)));
		QObject::connect(readingThread, SIGNAL(checkImagesUpdate(void)), this, SLOT(canImagesBeShowed(void)));
		QObject::connect(readingThread, SIGNAL(lastFrameRead()), this, SLOT(lastOriginalFramePushed()));
		QObject::connect(readingThread, SIGNAL(firstFrameRead()), this, SLOT(firstFrameAvailable()));

		QObject::connect(this, SIGNAL(stopReadingThread()), readingThread, SLOT(stopReading()));

		QObject::connect(refreshTimer, SIGNAL(timeout()), this, SLOT(setTimingIsOK()));

		readingThread->start();

		// FPS trigger
		refreshTimer->start(period);
	}
	else if(videoLoaded && !playing && paused) // Video is paused
	{
		paused = false;
		playing = true;
		refreshTimer->start(period);
	}
	refreshStatusBar();
}

// Called when pause button is clicked
void Cobi::on_pauseButton_clicked()
{
	if (videoLoaded && playing)  // Video is being played
	{
		playing = false;
		paused = true;
		refreshTimer->stop();
	}
	else if (videoLoaded && !playing && paused)  // Video is paused
	{
		playing = true;
		paused = false;
		refreshTimer->start(period);
	}
	refreshStatusBar();
}

// Called when stop button is clicked
void Cobi::on_stopButton_clicked()
{
	if (videoLoaded && (playing || paused))
	{
		clearReproduction(true);
	}
	refreshStatusBar();
}

// Stops threads and clears playing surfaces if clearPlayingSurface==TRUE
void Cobi::clearReproduction(bool clearPlayingSurface)
{
	// Stop reading thread
	emit stopReadingThread();
	readingThread->wait();
	delete readingThread;
	readingThread = NULL;

	// Stop processing thread
	emit stopProcessingThread();
	processingThread->wait();
	delete processingThread;
	processingThread = NULL;

	playing            = false;
	paused             = false;
	lastOriginalFrame  = false;
	lastProcessedFrame = false;

	videoLoaded = this->loadVideo();

	// Clear playing surfaces
	if(clearPlayingSurface)
	{
		ui.playSurface->makeCurrent();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		ui.playSurface->doneCurrent();
		ui.playSurface->updateGL();

		ui.playSurface2->makeCurrent();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		ui.playSurface2->doneCurrent();
		ui.playSurface2->updateGL();
	}

	// stop trigger
	refreshTimer->stop();
	ui.statusbar->showMessage("");

	if(savingOutput)
	{
		// stop writing thread
		writingThread->wait();
		delete writingThread;
		writingThread = NULL;
	}

	// clear image lists
	originalImage.clear();
	processInputImage.clear();
	processedImage.clear();
	outputImage.clear();

	ui.checkBox->setEnabled(true);
	this->on_checkBox_stateChanged();
}

// Updates the images shown in the playing surfaces
void Cobi::updateShowedImages()
{
	if(!originalImage.empty() && !processedImage.empty())
	{
		refreshStatusBar();

		// pop original image
		QMutexLocker lockOriginal(&originalImageMutex);
			CobiImage origIm = originalImage.front();
			originalImage.pop_front();
		lockOriginal.unlock();

		// pop processed image
		QMutexLocker lockProcessed(&processedImageMutex);
			CobiImage procIm = processedImage.front();
			processedImage.pop_front();
		lockProcessed.unlock();

		// refresh playing surfaces
		ui.playSurface->makeCurrent();
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, originalTextureName);
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
							origIm.getGLWidth(), origIm.getGLHeight(),
							GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid*)origIm.getGLData());
			glBegin(GL_TRIANGLE_STRIP);
				glTexCoord2f(0.0, 0.0);
				glVertex2f(0.0, 0.0);
				glTexCoord2f(1.0, 0.0);
				glVertex2f(1.0, 0.0);
				glTexCoord2f(0.0, 1.0);
				glVertex2f(0.0, 1.0);
				glTexCoord2f(1.0, 1.0);
				glVertex2f(1.0, 1.0);
			glEnd();
			glDisable(GL_TEXTURE_2D);
			glFlush();
		ui.playSurface->doneCurrent();
		ui.playSurface->updateGL();

		ui.playSurface2->makeCurrent();
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, processedTextureName);
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
							procIm.getGLWidth(), procIm.getGLHeight(),
							GL_RGBA, GL_UNSIGNED_BYTE, (GLvoid*)procIm.getGLData());
			glBegin(GL_TRIANGLE_STRIP);
				glTexCoord2f(0.0, 0.0);
				glVertex2f(0.0, 0.0);
				glTexCoord2f(1.0, 0.0);
				glVertex2f(1.0, 0.0);
				glTexCoord2f(0.0, 1.0);
				glVertex2f(0.0, 1.0);
				glTexCoord2f(1.0, 1.0);
				glVertex2f(1.0, 1.0);
			glEnd();
			glDisable(GL_TEXTURE_2D);
			glFlush();
		ui.playSurface2->doneCurrent();
		ui.playSurface2->updateGL();

		timingIsOK = false;
		refreshTimer->start(period);
	}
}

// Last frame of the input video read
void Cobi::lastOriginalFramePushed(void)
{
	lastOriginalFrame = true;
}

// Last frame of the input video processed
void Cobi::lastProcessedFramePushed(void)
{
	lastProcessedFrame = true;
}

// Tests if the playing surfaces can be refreshed
void Cobi::canImagesBeShowed(void)
{
	if (!originalImage.empty() && !processedImage.empty() && timingIsOK && playing)
	{
		updateShowedImages();
	}
	else if(originalImage.empty() && processedImage.empty() && (outputImage.empty() || !savingOutput) && lastProcessedFrame)
	{
		clearReproduction(true);
	}
}

// New image received
void Cobi::newReceivedImage(void)
{
	if(timingIsOK)
	{
		canImagesBeShowed();
	}
}

// Sets the timing as OK according to the framerate of the input video
void Cobi::setTimingIsOK()
{
	timingIsOK = true;
	canImagesBeShowed();
}

// Called when the first frame of the input video has been read
void Cobi::firstFrameAvailable()
{
	// launch processing thread and connect signals and slots
	processingThread = new ProcessVideoThread(this, &processInputImage, &processedImage, &outputImage, &bufferSize, &lastOriginalFrame, savingOutput,
											  &processInputImageMutex, &processedImageMutex, &outputImageMutex);
	QObject::connect(processingThread, SIGNAL(newPushedImage(void)), this, SLOT(newReceivedImage(void)));
	QObject::connect(processingThread, SIGNAL(checkImagesUpdate(void)), this, SLOT(canImagesBeShowed(void)));
	QObject::connect(processingThread, SIGNAL(lastFrameProcessed()), this, SLOT(lastProcessedFramePushed()));

	QObject::connect(this, SIGNAL(stopProcessingThread()), processingThread, SLOT(stopProcessing()));

	if(savingOutput)
	{
		// launch writing thread and connect signals and slots
		writingThread = new WriteVideoThread(this, &outputImage, &outputImageMutex, ui.lineEdit->text(), loadedVideoData->getCodecContext(), outputQuality);
		QObject::connect(writingThread, SIGNAL(errorWriting(void)), this, SLOT(errorWriting(void)));
		QObject::connect(this, SIGNAL(stopProcessingThread()), writingThread, SLOT(stopWriting()));
		writingThread->start();
	}
	processingThread->start();
}

// Error writing the output video
void Cobi::errorWriting(void)
{
	QString message("Error writing output video. The output will not be saved!");
	InfoWindow iwin(this, message);
    iwin.exec();
}

// Refreshes the text in the status bar
void Cobi::refreshStatusBar(void)
{

    if(playing)
    {
    	if(processingThread!=NULL && processingThread->isRunning())
    	{
    		ui.statusbar->showMessage(QString("PLAYING | FPS (input): %1 | FPS (processing): %2 | ").arg(fps).arg(QString().setNum(processingThread->getFPS(),'f',2)));
    	}
    	else
    	{
    		ui.statusbar->showMessage(QString("PLAYING | FPS (input): %1 | ").arg(fps));
    	}
    }
    else if(paused)
    {
    	ui.statusbar->showMessage(QString("PAUSED | "));
    }
    else if(videoLoaded)
    {
    	ui.statusbar->showMessage(QString("VIDEO LOADED | "));
    }
    else
    {
    	ui.statusbar->showMessage(QString(""));
    }

	ui.statusbar->showMessage(ui.statusbar->currentMessage().append(QString(" RB: %1/%2 - IB: %3/%2 - PB: %4/%2 - OB: %5/%2").arg(originalImage.size()).arg(bufferSize).arg(processInputImage.size()).arg(processedImage.size()).arg(outputImage.size())));
}
