/*
 * Copyright (C) 2010 EPFL (Ecole Polytechnique fédérale de Lausanne)
 * Transport and Mobility Laboratory (Transp-OR)
 *
 * This file is part of COBI.
 *
 * COBI is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * COBI is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * COBI; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA.
 *
 * Author: Javier Cruz-Mota
 *         javier.cruz{AT}epfl.ch
 *         http://transp-or2.epfl.ch/pagesPerso/javier.php
 *
 */

#include "CobiImage.h"

// Constructor
CobiImage::CobiImage()
{
	imageLoaded = false;
	useGL       = false;
	useOCV      = false;
}

// Constructor
//   Parameters:
//      ci -> reference to a CobiImage that will be copied
CobiImage::CobiImage(const CobiImage& ci)
{
	this->image    = QImage(*(ci.getQImage())).copy(QRect(0,0,ci.getQWidth(),ci.getQHeight()));

	this->useGL = ci.isGLUsed();
	if(this->useGL)
	{
		this->glImage  = QImage(*(ci.getGLImage())).copy(QRect(0,0,ci.getGLWidth(),ci.getGLHeight()));
	}

	this->useOCV = ci.isOCVUsed();
	if(this->useOCV)
	{
		this->ocvImage = cvCloneImage(ci.getOCVImage());
	}

	this->imageLoaded = true;
}

// Constructor
//   Parameters:
//      data   -> buffer with pixel data
//      width  -> image width
//      height -> image height
//      gl     -> true if OpenGL image has to be initialised
//      ocv    -> true if OpenCV image has to be initialised
CobiImage::CobiImage(const unsigned char *data, int width, int height, bool gl, bool ocv)
{
	image = QImage(data, width, height, QImage::Format_RGB32).copy(QRect(0,0,width,height));

	useGL = gl;
	if(gl)
	{
		glImage = QImage(QGLWidget::convertToGLFormat(image));
	}

	useOCV = ocv;
	if(ocv)
	{
		ocvImage = cvCreateImage(cvSize(width,height), IPL_DEPTH_8U, 4);
		memcpy(ocvImage->imageData, data, width*height*4*sizeof(unsigned char));
	}

	imageLoaded = true;
}

// "=" operator for CobiImage
CobiImage& CobiImage::operator=(const CobiImage &pI)
{
	if(this->imageLoaded && this->useOCV)
	{
		cvReleaseImage(&this->ocvImage);
	}

	this->image   = QImage(*pI.getQImage()).copy(QRect(0,0,pI.getQWidth(),pI.getQHeight()));

	this->useGL = pI.isGLUsed();
	if(this->useGL)
	{
		this->glImage = QImage(*pI.getGLImage()).copy(QRect(0,0,pI.getGLWidth(),pI.getGLHeight()));
	}

	this->useOCV = pI.isOCVUsed();
	if(this->useOCV)
	{
		this->ocvImage = cvCloneImage(pI.getOCVImage());
	}

	this->imageLoaded = true;

	return *this;
}

// Destructor
CobiImage::~CobiImage()
{
	if(imageLoaded)
	{
		if(useOCV)
		{
			cvReleaseImage(&this->ocvImage);
		}

		imageLoaded = false;
		useGL       = false;
		useOCV      = false;
	}
}

// Loads an image of size width*height from a given buffer buffer
void CobiImage::loadImage(const unsigned char *buffer, int width, int height, bool gl, bool ocv)
{
	if(imageLoaded && useOCV)
	{
		cvReleaseImage(&this->ocvImage);
	}
	image   = QImage(buffer, width, height, QImage::Format_RGB32).copy(QRect(0,0,width,height));

	useGL = gl;
	if(gl)
	{
		glImage = QImage(QGLWidget::convertToGLFormat(image)).copy(QRect(0,0,width,height));
	}

	useOCV = ocv;
	if(ocv)
	{
		ocvImage = cvCreateImage(cvSize(width,height), IPL_DEPTH_8U, 4);
		memcpy(ocvImage->imageData, buffer, width*height*4*sizeof(unsigned char));
	}

	imageLoaded = true;
}

// Recomputes the QImage in OpenGL format
void CobiImage::refreshGLImage(void)
{
	if(imageLoaded)
	{
		glImage = QImage(QGLWidget::convertToGLFormat(image)).copy(QRect(0,0,image.width(),image.height()));
		useGL = true;
	}
}

// Refreshes the OpenCV IplImage
void CobiImage::refreshOCVImage(void)
{
	if(imageLoaded)
	{
		if(useOCV)
		{
			cvReleaseImage(&this->ocvImage);
		}
		int width = image.width();
		int height = image.height();
		ocvImage = cvCreateImage(cvSize(width,height), IPL_DEPTH_8U, 4);
		memcpy(ocvImage->imageData, image.bits(), width*height*4*sizeof(unsigned char));
		useOCV = true;
	}
}

// Loads GL image if it is not already loaded
void CobiImage::loadGLImage(void)
{
	if(imageLoaded && !useGL)
	{
		glImage = QImage(QGLWidget::convertToGLFormat(image)).copy(QRect(0,0,image.width(),image.height()));
		useGL = true;
	}
}

// Loads OCV image if it is not already loaded
void CobiImage::loadOCVImage(void)
{
	if(imageLoaded && !useOCV)
	{
		int width = image.width();
		int height = image.height();

		ocvImage = cvCreateImage(cvSize(width,height), IPL_DEPTH_8U, 4);
		memcpy(ocvImage->imageData, image.bits(), width*height*4*sizeof(unsigned char));
		useOCV = true;
	}
}

// Returns a pointer to the image
QImage* CobiImage::getQImage()
{
	if(imageLoaded)
	{
		return &image;
	}
	else
	{
		return NULL;
	}
}

// Returns a const pointer to the image
const QImage* CobiImage::getQImage() const
{
	if(imageLoaded)
	{
		return &image;
	}
	else
	{
		return NULL;
	}
}

// Returns a pointer to the GL image
QImage* CobiImage::getGLImage()
{
	if(imageLoaded)
	{
		if(!useGL)
		{
			this->loadGLImage();
		}
		return &glImage;
	}
	else
	{
		return NULL;
	}
}

// Returns a const pointer to the GL image
const QImage* CobiImage::getGLImage() const
{
	if(imageLoaded && useGL)
	{
		return &glImage;
	}
	else
	{
		return NULL;
	}
}

// Returns a const pointer to the OpenCV image
const IplImage* CobiImage::getOCVImage() const
{
	if(imageLoaded && useOCV)
	{
		return ocvImage;
	}
	else
	{
		return NULL;
	}
}

// Returns a pointer to the OpenCV image
IplImage* CobiImage::getOCVImage()
{
	if(imageLoaded)
	{
		if(!useOCV)
		{
			this->loadOCVImage();
		}
		return ocvImage;
	}
	else
	{
		return NULL;
	}
}

// Returns image width
int CobiImage::getQWidth(void) const
{
	if(imageLoaded)
	{
		return image.width();
	}
	else
	{
		return 0;
	}
}

// Returns image height
int CobiImage::getQHeight(void) const
{
	if(imageLoaded)
	{
		return image.height();
	}
	else
	{
		return 0;
	}
}

// Returns GL image width
int CobiImage::getGLWidth(void) const
{
	if(imageLoaded && useGL)
	{
		return glImage.width();
	}
	else
	{
		return 0;
	}
}

// Returns GL image width
int CobiImage::getGLWidth(void)
{
	if(imageLoaded)
	{
		if(!useGL)
		{
			this->loadGLImage();
		}
		return glImage.width();
	}
	else
	{
		return 0;
	}
}

// Returns GL image height
int CobiImage::getGLHeight(void) const
{
	if(imageLoaded && useGL)
	{
		return glImage.height();
	}
	else
	{
		return 0;
	}
}

// Returns GL image height
int CobiImage::getGLHeight(void)
{
	if(imageLoaded)
	{
		if(!useGL)
		{
			this->loadGLImage();
		}
		return glImage.height();
	}
	else
	{
		return 0;
	}
}

// Returns OpenCV image width
int CobiImage::getOCVWidth(void) const
{
	if(imageLoaded && useOCV)
	{
		return ocvImage->width;
	}
	else
	{
		return 0;
	}
}

// Returns OpenCV image width
int CobiImage::getOCVWidth(void)
{
	if(imageLoaded)
	{
		if(!useOCV)
		{
			this->loadOCVImage();
		}
		return ocvImage->width;
	}
	else
	{
		return 0;
	}
}

// Returns OpenCV image height
int CobiImage::getOCVHeight(void) const
{
	if(imageLoaded && useOCV)
	{
		return ocvImage->height;
	}
	else
	{
		return 0;
	}
}

// Returns OpenCV image height
int CobiImage::getOCVHeight(void)
{
	if(imageLoaded)
	{
		if(!useOCV)
		{
			this->loadOCVImage();
		}
		return ocvImage->height;
	}
	else
	{
		return 0;
	}
}

// Returns a pointer to the data of the image
unsigned char* CobiImage::getQData(void)
{
	if(imageLoaded)
	{
		return image.bits();
	}
	else
	{
		return NULL;
	}
}

// Returns a pointer to the data of the GL image
unsigned char* CobiImage::getGLData(void)
{
	if(imageLoaded)
	{
		if(!useGL)
		{
			this->loadGLImage();
		}
		return glImage.bits();
	}
	else
	{
		return NULL;
	}
}

// Returns a pointer to the data of the OpenCV image
char* CobiImage::getOCVData(void)
{
	if(imageLoaded)
	{
		if(!useOCV)
		{
			this->loadOCVImage();
		}
		return ocvImage->imageData;
	}
	else
	{
		return NULL;
	}
}

// Returns true if GL image is being used
bool CobiImage::isGLUsed(void) const
{
	return useGL;
}

// Returns true if IplImage (OpenCV) is being used
bool CobiImage::isOCVUsed(void) const
{
	return useOCV;
}

// Converts IplImage to QImage
// Code adapted from the code published by Patrik08 in:
// http://www.qtcentre.org/forum/f-qt-programming-2/t-opencv-integration-11655.html
// This version only supports IPL_DEPTH_8U as input depth (with 1 or 3 channels) and RGB32 as QImage format
void CobiImage::refreshQFromOCV(void)
{
	int width     = ocvImage->width;
	int widthStep = ocvImage->widthStep;
	int height    = ocvImage->height;

	unsigned char *qImageBuffer = (unsigned char *)malloc(width * height * 4 * sizeof(unsigned char));;
    unsigned char *qImagePointer = qImageBuffer;
    const unsigned char *iplImagePointer = (const unsigned char *)ocvImage->imageData;

	switch (ocvImage->depth)
	{
		case IPL_DEPTH_8U:
			if (ocvImage->nChannels == 1)
			{
				for (int y = 0; y < height; ++y)
				{
					for (int x = 0; x < width; ++x)
					{
						qImagePointer[0] = iplImagePointer[0];
						qImagePointer[1] = iplImagePointer[0];
						qImagePointer[2] = iplImagePointer[0];
						qImagePointer[3] = 0;

						qImagePointer += 4;
						iplImagePointer++;
					}
					iplImagePointer += widthStep - width;
				}
			}
			else if (ocvImage->nChannels == 3)
			{
				for (int y = 0; y < height; ++y)
				{
					for (int x = 0; x < width; ++x)
					{
						qImagePointer[0] = iplImagePointer[0];
						qImagePointer[1] = iplImagePointer[1];
						qImagePointer[2] = iplImagePointer[2];
						qImagePointer[3] = 0;

						qImagePointer += 4;
						iplImagePointer += 3;
					}
					iplImagePointer += widthStep - 3 * width;
				}
			}
			else
			{
				qDebug("IplImageToQImage: image format is not supported : depth=8U and %d channels\n", ocvImage->nChannels);
			}
			break;

		default:
			qDebug("IplImageToQImage: image format is not supported : depth=%d and %d channels\n", ocvImage->depth, ocvImage->nChannels);
			break;
	}
	image = QImage(qImageBuffer, width, height, QImage::Format_RGB32).copy(QRect(0,0,width,height));
	free(qImageBuffer);
}
