/*
 * Copyright (C) 2010 EPFL (Ecole Polytechnique fédérale de Lausanne)
 * Transport and Mobility Laboratory (Transp-OR)
 *
 * This file is part of COBI.
 *
 * COBI is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * COBI is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * COBI; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA.
 *
 * Author: Javier Cruz-Mota
 *         javier.cruz{AT}epfl.ch
 *         http://transp-or2.epfl.ch/pagesPerso/javier.php
 *
 */

#include "ProcessVideoThread.h"

// Constructor
//   Parameters:
//      parent -> pointer to the parent QObject
//      ii     -> pointer to a CobiImageList object (input frames for the processing thread)
//      pi     -> pointer to a CobiImageList object (output frames of the processing thread)
//      oi     -> pointer to a CobiImageList object (input frames for the writing thread)
//      ls     -> pointer to an integer containing the size of the lists
//      lf     -> pointer to an boolean (true when the last frame has been read)
//      so     -> boolean (true => output is saved by a writing thread)
//      im     -> pointer to a QMutex for accessing the "ii" list
//      pm     -> pointer to a QMutex for accessing the "pi" list
//      om     -> pointer to a QMutex for accessing the "oi" list
ProcessVideoThread::ProcessVideoThread(QObject *parent, CobiImageList *ii, CobiImageList *pi, CobiImageList *oi, int *ls, bool *lf, bool so, QMutex *im, QMutex *pm, QMutex *om)
: QThread(parent)
{
	inputImage             = ii;
	processedImage         = pi;
	outputImage			   = oi;
	listSize               = ls;
	lastFrameRead          = lf;
	savingOutput		   = so;
	processInputImageMutex = im;
	processedImageMutex    = pm;
	outputImageMutex	   = om;
	stopLoop			   = false;
}

// Destructor
ProcessVideoThread::~ProcessVideoThread()
{

}

// Main loop of the processing thread
void ProcessVideoThread::run()
{
	elapsedTime.start();

	while((!*lastFrameRead || !inputImage->empty()) && !stopLoop)
	{
		if(inputImage->empty() || (int) outputImage->size()>= *listSize)
		{
			this->usleep(50);
		}
		else
		{
			if(!stopLoop)
			{
				// Pop the front image from the list of input frames
				QMutexLocker lockerInput(processInputImageMutex);
				currentImage = inputImage->front();
				inputImage->pop_front();
				lockerInput.unlock();
			}

			// Process currentImage
			elapsedTime.restart();
			this->process();
			elapsedTimeValue = elapsedTime.elapsed();

			// While the list is full, wait
			while(((int) processedImage->size()>= *listSize ) && !stopLoop)
			{
				emit checkImagesUpdate();
				this->usleep(50);
			}


			if(!stopLoop)
			{
				// Push the processed image into the list of processed frames
				QMutexLocker lockerProcessed(processedImageMutex);
				processedImage->push_back(currentImage);
				lockerProcessed.unlock();

				// If the output is saved, push the processed image into the list
				// of output frames
				if(savingOutput)
				{
					while(((int) outputImage->size()>= *listSize) && !stopLoop)
					{
						this->usleep(50);
					}
					if(!stopLoop)
					{
						QMutexLocker lockerOutput(outputImageMutex);
						outputImage->push_back(currentImage);
						lockerOutput.unlock();
					}
				}

				emit newPushedImage();
			}
		}
	}
	if(!stopLoop)
	{
		emit lastFrameProcessed();
	}
	this->exit();
}

// Forces the end of the main loop
void ProcessVideoThread::stopProcessing(void)
{
	stopLoop = true;
}

// Method that performs the processing of each frame
void ProcessVideoThread::process(void)
{
	// Example of frame processing. It takes the current frame,
	//  converts it to grayscale and applies a Gaussian smoothing with
	//  sigma=17. Then, it refreshes the QImage data using the obtained
	//  smoothed image and releases the temporal IplImage.
	IplImage *tmpImage=cvCloneImage(currentImage.getOCVImage());

	currentImage.getOCVImage()->nChannels = 1;
	cvCvtColor(tmpImage, currentImage.getOCVImage(), CV_RGBA2GRAY);
	cvSmooth(currentImage.getOCVImage(), currentImage.getOCVImage(), CV_GAUSSIAN, 17);
	currentImage.refreshQFromOCV();

	cvReleaseImage(&tmpImage);

	// [DEBUG] Multithread testing:
	//this->msleep(int(0.5+1000.0*rand()/float(RAND_MAX)));
}

// Returns the processing time
int ProcessVideoThread::getElapsedTime(void)
{
	return elapsedTimeValue;
}

// Returns the FPS of the processing
double ProcessVideoThread::getFPS(void)
{
	return 1000.0/elapsedTimeValue;
}
