/*
 * Copyright (C) 2010 EPFL (Ecole Polytechnique fédérale de Lausanne)
 * Transport and Mobility Laboratory (Transp-OR)
 *
 * This file is part of COBI.
 *
 * COBI is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * COBI is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * COBI; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA.
 *
 * Author: Javier Cruz-Mota
 *         javier.cruz{AT}epfl.ch
 *         http://transp-or2.epfl.ch/pagesPerso/javier.php
 *
 */

#include "ReadVideoThread.h"

// Constructor
//   Parameters:
//      parent -> pointer to the parent QObject
//      vd     -> pointer to the VideoData object containing the data of the read video
//      oi     -> pointer to a CobiImageList object (frames of the input video)
//      ii     -> pointer to a CobiImageList object (input frames for the processing thread)
//      ls     -> pointer to an integer containing the size of the lists
//      om     -> pointer to a QMutex for accessing the "oi" list
//      im     -> pointer to a QMutex for accessing the "ii" list
ReadVideoThread::ReadVideoThread(QObject *parent, VideoData *vd, CobiImageList *oi, CobiImageList *ii, int *ls, QMutex *om, QMutex *im)
: QThread(parent)
{
	loadedVideoData        = vd;
	originalImage          = oi;
	inputImage             = ii;
	listSize               = ls;
	originalImageMutex     = om;
	processInputImageMutex = im;
	stopLoop	       = false;
}

// Run loop of the thread
void ReadVideoThread::run()
{
	AVPacket packet;
	int state;
	bool firstFrame = true;

	state = loadedVideoData->ReadFrame(&packet);

	while(state != VideoData::LASTVIDEOFRAME && !stopLoop)
	{
		if(state == VideoData::VIDEOFRAME)
		{
			if(firstFrame)
			{
				emit firstFrameRead();
				firstFrame = false;
			}

			// Convert the frame to RGB format
			loadedVideoData->ConvertToRGB();

			// While the list is full, wait
			while((int) originalImage->size()>= *listSize && !stopLoop)
			{
				emit checkImagesUpdate();
				this->usleep(50);
			}

			if(!stopLoop)
			{
				// Push the image into the originalImage list
				QMutexLocker lockerOriginal(originalImageMutex);
				originalImage->push_back(CobiImage( loadedVideoData->getRGBFrameData(), loadedVideoData->getWidth(), loadedVideoData->getHeight(), false, false));
				lockerOriginal.unlock();
			}

			// While the list is full, wait
			while((int) inputImage->size()>= *listSize && !stopLoop)
			{
				this->usleep(50);
			}
			if(!stopLoop)
			{
				// Push the image into the inputImage list
				QMutexLocker lockerInput(processInputImageMutex);
				inputImage->push_back(CobiImage( loadedVideoData->getRGBFrameData(), loadedVideoData->getWidth(), loadedVideoData->getHeight(), false, false));
				lockerInput.unlock();
				emit newPushedImage();
			}
		}
		if(!stopLoop)
		{
			// Read another frame
			state = loadedVideoData->ReadFrame(&packet);

			// [DEBUG] Multithread testing:
			//this->msleep(int(0.5+1000.0*rand()/float(RAND_MAX)));
		}
	}
	av_free_packet(&packet);
	if(!stopLoop)
	{
		emit lastFrameRead();
	}
	this->exit();
}

// Forces exit of the main loop in ReadVideoThread::run
void ReadVideoThread::stopReading(void)
{
	stopLoop = true;
}
