/*
 * Copyright (C) 2010 EPFL (Ecole Polytechnique fédérale de Lausanne)
 * Transport and Mobility Laboratory (Transp-OR)
 *
 * This file is part of COBI.
 *
 * COBI is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * COBI is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * COBI; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA.
 *
 * Author: Javier Cruz-Mota
 *         javier.cruz{AT}epfl.ch
 *         http://transp-or2.epfl.ch/pagesPerso/javier.php
 *
 */

#include "VideoData.h"

// Constructor
VideoData::VideoData()
{
	videoDataIsLoaded = false;
	pFormatCtx     = NULL;
	pCodecCtx      = NULL;
	pCodec         = NULL;
	pFrame         = NULL;
	pFrameRGB      = NULL;
	imageBuffer    = NULL;
	imageRGBBuffer = NULL;
	convertContext = NULL;
}

// Destructor
VideoData::~VideoData()
{
	if (videoDataIsLoaded)
	{
		sws_freeContext(convertContext);
		avcodec_close(pCodecCtx);
		av_close_input_file(pFormatCtx);
		av_free(imageBuffer);
		av_free(imageRGBBuffer);
		av_free(pFrame);
		av_free(pFrameRGB);
	}
}

// Reads headers of the video file "fileName"
bool VideoData::LoadData(const char *fileName)
{
        avcodec_register_all();
        av_register_all();
	if( av_open_input_file(&pFormatCtx, fileName, NULL, 0, NULL) == 0 )
	{
		if(av_find_stream_info(pFormatCtx)>=0)
		{
			unsigned int i = 0;
			videoStream = -1;
			while (i<pFormatCtx->nb_streams && videoStream==-1)
			{
				if (pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO)
				{
					videoStream=i;
				}
			}
			if (videoStream!=-1)
			{
				pCodecCtx = pFormatCtx->streams[videoStream]->codec;
				pCodec = avcodec_find_decoder(pCodecCtx->codec_id);
    			if(pCodec!=NULL)
    			{
        			if (avcodec_open(pCodecCtx, pCodec)==0)
        			{
        				pFrame = avcodec_alloc_frame();
        				pFrameRGB = avcodec_alloc_frame();
        				imageBuffer = (quint8 *) av_malloc(avpicture_get_size(PIX_FMT_RGB32,
														   pCodecCtx->width,
														   pCodecCtx->height)*sizeof(quint8));
        				imageRGBBuffer = (quint8 *) av_malloc(avpicture_get_size(PIX_FMT_RGB32,
															pCodecCtx->width,
        													pCodecCtx->height)*sizeof(quint8));
        				avpicture_fill((AVPicture *) pFrame, imageBuffer,
									   pCodecCtx->pix_fmt, pCodecCtx->width,
									   pCodecCtx->height);
        				avpicture_fill((AVPicture *) pFrameRGB, imageRGBBuffer,
									   PIX_FMT_RGB32, pCodecCtx->width,
									   pCodecCtx->height);
        				convertContext = sws_getContext(pCodecCtx->width, pCodecCtx->height,
        										    pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height,
        										    PIX_FMT_RGB32, SWS_FAST_BILINEAR, NULL, NULL, NULL);
						videoDataIsLoaded = true;
        			}
    			}
			}
		}
	}
	return videoDataIsLoaded;
}

// Reads a frame
int VideoData::ReadFrame(AVPacket *packet)
{
	int frameFinished;

	if ( av_read_frame(pFormatCtx, packet)>=0 )
	{
		if (packet->stream_index == videoStream) {
			avcodec_decode_video2(pCodecCtx, pFrame, &frameFinished, packet);
			return VIDEOFRAME;
		}
		return NOVIDEOFRAME;
	}
	return LASTVIDEOFRAME;
}

// Converts a frame to RGB format
void VideoData::ConvertToRGB(void)
{
	sws_scale(convertContext, pFrame->data, pFrame->linesize, 0, pCodecCtx->height,
			  pFrameRGB->data, pFrameRGB->linesize);
}

// Returns video frames width
int VideoData::getWidth(void)
{
	return pCodecCtx->width;
}

// Returns video frames height
int VideoData::getHeight(void)
{
	return pCodecCtx->height;
}

// Pointer to the data of a frame
uint8_t* VideoData::getFrameData(void)
{
	return pFrame->data[0];
}

// Pointer to the RGB data of a frame
uint8_t* VideoData::getRGBFrameData(void)
{
	return pFrameRGB->data[0];
}

// Time separation between frames in milliseconds
double VideoData::getPeriod(void)
{
	double period = 1000*((double) pCodecCtx->time_base.num)/((double) pCodecCtx->time_base.den);
	if(period>=1)
	{
		return period;
	}
	else
	{
		return 1000*period;
	}
}

// Returns the codec context
AVCodecContext* VideoData::getCodecContext(void)
{
	return pCodecCtx;
}

