/*
 * Copyright (C) 2010 EPFL (Ecole Polytechnique fédérale de Lausanne)
 * Transport and Mobility Laboratory (Transp-OR)
 *
 * This file is part of COBI.
 *
 * COBI is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * COBI is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * COBI; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA 02111-1307 USA.
 *
 * Author: Javier Cruz-Mota
 *         javier.cruz{AT}epfl.ch
 *         http://transp-or2.epfl.ch/pagesPerso/javier.php
 *
 */

#include "WriteVideoThread.h"

// Constructor
//   Parameters:
//      parent -> pointer to the parent QObject
//      oi     -> pointer to a CobiImageList object (frames of the input video)
//      om     -> pointer to a QMutex for accessing the "oi" list
//      fn     -> QString containing the file name to save
//      ic     -> pointer to the codec context of the input video
//      oq     -> quality of the output video
WriteVideoThread::WriteVideoThread(QObject *parent, CobiImageList *oi, QMutex *om, QString fn, AVCodecContext *ic, int oq)
: QThread(parent)
{
	stopLoop          = false;
	outputImage       = oi;
	outputImageMutex  = om;
	fileName          = fn;
	videoStream       = NULL;
	videoBuffer       = NULL;
	inputCodecContext = ic;
	videoBufferSize   = 200000;
	tmpPicture        = NULL;
	frameCount	  = 0;
	convertContext	  = NULL;
	outputQuality     = oq;
}

// Destructor
WriteVideoThread::~WriteVideoThread()
{
}

// Prepares a video stream for the writing
void WriteVideoThread::addVideoStream(void)
{
    videoStream = av_new_stream(outputFormatContext, 0);
    if (!videoStream)
    {
    	emit errorWriting();
    	this->exit(1);
    }
    outputCodecContext = videoStream->codec;

    outputCodecContext->codec_id      = outputFormat->video_codec;
    outputCodecContext->codec_type    = CODEC_TYPE_VIDEO;
    outputCodecContext->bit_rate      = inputCodecContext->bit_rate;
    outputCodecContext->qmin          = 2;
    outputCodecContext->qmax          = outputQuality;
    outputCodecContext->max_qdiff     = 2;
    outputCodecContext->width         = inputCodecContext->width;
    outputCodecContext->height        = inputCodecContext->height;
    outputCodecContext->time_base.den = inputCodecContext->time_base.den;
    outputCodecContext->time_base.num = inputCodecContext->time_base.num;
    outputCodecContext->gop_size      = inputCodecContext->gop_size;
    outputCodecContext->pix_fmt       = PIX_FMT_YUV420P;
}

// Allocates a picture where store frames
bool WriteVideoThread::allocPicture(void)
{
	int size;

	// YUV picture
    this->picture = avcodec_alloc_frame();
    if (!this->picture)
    {
        return false;
    }
    size = avpicture_get_size(videoStream->codec->pix_fmt, videoStream->codec->width, videoStream->codec->height);
    this->pictureBuffer = (uint8_t *) av_malloc(size);
    if (!this->pictureBuffer)
    {
        av_free(this->picture);
        return false;
    }
    avpicture_fill((AVPicture *) this->picture, this->pictureBuffer, videoStream->codec->pix_fmt, videoStream->codec->width, videoStream->codec->height);

    // RGB picture
	this->tmpPicture = avcodec_alloc_frame();
	if (!this->tmpPicture)
	{
		return false;
	}
	size = avpicture_get_size(PIX_FMT_RGB32, videoStream->codec->width, videoStream->codec->height);
	this->tmpPictureBuffer = (uint8_t *) av_malloc(size);
	if (!this->tmpPictureBuffer)
	{
		av_free(this->tmpPicture);
		return false;
	}
	avpicture_fill((AVPicture *)this->tmpPicture, this->tmpPictureBuffer, PIX_FMT_RGB32, videoStream->codec->width, videoStream->codec->height);

    return true;
}

// Opens the video file
void WriteVideoThread::openVideo(void)
{
    AVCodec *codec;

    codec = avcodec_find_encoder(videoStream->codec->codec_id);
    if (!codec)
    {
    	emit errorWriting();
    	this->exit(1);
    }

    if (avcodec_open(videoStream->codec, codec) < 0)
    {
    	emit errorWriting();
    	this->exit(1);
    }

    if(!(outputFormatContext->oformat->flags & AVFMT_RAWPICTURE))
    {
        videoBuffer = (uint8_t *) av_malloc(videoBufferSize);
    }

    if (!this->allocPicture())
    {
    	emit errorWriting();
    	this->exit(1);
    }
}

// Copies the data from a CobiImage to an AVPicture
void WriteVideoThread::fillImage(void)
{
    memcpy(tmpPicture->data[0], currentImage.getQData(), currentImage.getQWidth()*currentImage.getQHeight()*4*sizeof(unsigned char));
}

// Writes a frame to the video
void WriteVideoThread::writeVideoFrame(void)
{
    int outSize;
    int status;

	if (convertContext == NULL)
	{
		convertContext = sws_getContext(outputCodecContext->width, outputCodecContext->height,
									   PIX_FMT_RGB32,
									   outputCodecContext->width, outputCodecContext->height,
									   PIX_FMT_YUV420P,
									   SWS_FAST_BILINEAR, NULL, NULL, NULL);

		if (convertContext == NULL)
		{
	    	emit errorWriting();
	    	this->exit(1);
		}
	}
	this->fillImage();
	sws_scale(convertContext, tmpPicture->data, tmpPicture->linesize,
			  0, outputCodecContext->height, picture->data, picture->linesize);

    if (outputFormatContext->oformat->flags & AVFMT_RAWPICTURE)
    {
        AVPacket pkt;
        av_init_packet(&pkt);

        pkt.flags |= PKT_FLAG_KEY;
        pkt.stream_index = videoStream->index;
        pkt.data = (uint8_t *)this->picture;
        pkt.size = sizeof(AVPicture);

        status = av_write_frame(outputFormatContext, &pkt);
	av_free_packet(&pkt);
    }
    else
    {
        outSize = avcodec_encode_video(outputCodecContext, videoBuffer, videoBufferSize, this->picture);
        if (outSize > 0)
        {
		AVPacket pkt;
		av_init_packet(&pkt);

		pkt.pts= av_rescale_q(outputCodecContext->coded_frame->pts, outputCodecContext->time_base, videoStream->time_base);
		if(outputCodecContext->coded_frame->key_frame)
		{
			pkt.flags |= PKT_FLAG_KEY;
		}
		pkt.stream_index= videoStream->index;
		pkt.data= videoBuffer;
		pkt.size= outSize;

		status = av_write_frame(outputFormatContext, &pkt);
		av_free_packet(&pkt);
        }
        else
        {
            status = 0;
        }
    }
    if (status != 0)
    {
    	emit errorWriting();
    	this->exit(1);
    }
    frameCount++;
}

// Closes the output file
void WriteVideoThread::closeVideo(void)
{
    av_free(picture->data[0]);
    av_free(picture);
    av_free(tmpPicture->data[0]);
    av_free(tmpPicture);
    av_free(videoBuffer);
    sws_freeContext(convertContext);
    avcodec_close(videoStream->codec);
}

// Main loop of the thread
void WriteVideoThread::run(void)
{
	outputFormat = av_guess_format(NULL, fileName.toStdString().c_str(), NULL);
	if (!outputFormat)
	{
		outputFormat = av_guess_format("avi", NULL, NULL);
		if (!outputFormat)
		{
			emit errorWriting();
			this->exit(1);
		}
	}

	outputFormatContext = avformat_alloc_context();
	if (!outputFormatContext)
	{
		emit errorWriting();
		this->exit(1);
	}
	outputFormatContext->oformat = outputFormat;
	snprintf(outputFormatContext->filename, sizeof(outputFormatContext->filename), "%s", fileName.toStdString().c_str());

	if (outputFormat->video_codec != CODEC_ID_NONE)
	{
		addVideoStream();
	}

	if (av_set_parameters(outputFormatContext, NULL) < 0)
	{
		emit errorWriting();
		this->exit(1);
	}

	dump_format(outputFormatContext, 0, fileName.toStdString().c_str(), 1);

	if (videoStream)
	{
		openVideo();
	}

	if (!(outputFormat->flags & AVFMT_NOFILE))
	{
		if (url_fopen(&outputFormatContext->pb, fileName.toStdString().c_str(), URL_WRONLY) < 0)
		{
			emit errorWriting();
			this->exit(1);
		}
	}

	av_write_header(outputFormatContext);

	while(!stopLoop)
	{
		while(outputImage->empty() && !stopLoop)
		{
			this->usleep(50);
		}
		if(!stopLoop)
		{
			QMutexLocker lockerOutput(outputImageMutex);
				currentImage = outputImage->front();
				outputImage->pop_front();
			lockerOutput.unlock();

			// [DEBUG] Multithread testing:
			//this->msleep(int(0.5+1000.0*rand()/float(RAND_MAX)));
		}

		if (videoStream && !stopLoop)
		{
			this->writeVideoFrame();
		}
	}

	if (videoStream)
	{
		this->closeVideo();
	}

	av_write_trailer(outputFormatContext);

	for(unsigned int i = 0; i < outputFormatContext->nb_streams; i++)
	{
		av_freep(&outputFormatContext->streams[i]->codec);
		av_freep(&outputFormatContext->streams[i]);
	}

	if (!(outputFormat->flags & AVFMT_NOFILE))
	{
		url_fclose(outputFormatContext->pb);
	}

	av_free(outputFormatContext);

	this->exit();
}

// Forces the end of the main loop
void WriteVideoThread::stopWriting(void)
{
	stopLoop = true;
}
